/* ********************************************************************
 * Copyright (c) 2021 Biological Modeling and Visualization Laboratory,
 * University of Calgary. All rights reserved.
 * ********************************************************************/

/* Descriptive model of Canola (Brassica napus L.) */
#include <algorithm>
#include <cmath>
#include <lpfgall.h>

/* Declaration of constants */

#define DT 1. // simulation time step
#define STEPS int(val(MaxPlantAge) / DT);

#define DX 0.05 // step size for the leaf geometry
#define DX2 0.1 // step size for the petal geometry

/* Mapping of functions from functions.fset and functions.tset to macros */

#define branch_angle(n)     func(BRANCH_ANGLE,n/(val(MaxPhytomers)-val(VegPhytomers)))
#define branch_tropism(n)   func(BRANCH_TROPISM,n/val(VegPhytomers))

#define int_length(t)       tfunc(INT_LENGTH,t)
#define int_target_len(n)   func(INT_TARGET_LEN,n/val(MaxPhytomers))

#define leaf_length(t)      tfunc(LEAF_LENGTH,t)
#define leaf_target_len(n)  func(LEAF_TARGET_LEN,n/val(VegPhytomers))
#define leaf_target_wid(n)  func(LEAF_WIDTH_POS,n/val(VegPhytomers)) // NOTE: this function was set to 1 in the paper
#define leaf_inclination(t) tfunc(LEAF_INCLINATION,t) // NOTE: this function was set to 1 in the paper
#define leaf_width(x)       func(LEAF_WIDTH,x)
#define leaf_bend(x)        (360.0*func(LEAF_BEND,x))
#define leaf_angle(n)       func(LEAF_ANGLE,n/val(VegPhytomers))

#define pedicel_length(t)   tfunc(PEDICEL_LENGTH,t)

#define petal_angle(t)      tfunc(PETAL_ANGLE,t)
#define petal_length(t)	    tfunc(PETAL_LENGTH,t)
#define petal_width(x)      func(PETAL_WIDTH,x)

#define pod_length(t)       tfunc(POD_LENGTH,t)
#define prob_podding(t)     tfunc(PROB_PODDING,t/val(MaxPlantAge))

#define bud_vigour(n)       func(BUD_VIGOUR,n/val(VegPhytomers))
#define bud_delay(n)        func(BUD_DELAY,n/val(VegPhytomers))
#define bud_init_node(n)    func(BUD_INIT_NODE,n/val(VegPhytomers))

#define plastochron(n)      func(PLASTOCHRON,n/val(VegPhytomers))

/* Structure declarations and initialization */

struct OrganData {
	float age;
	float n; // NOTE: changed to 'float' from 'int' to avoid type casting 
	int order;
	int state;
    // set default values
	OrganData()
	  : age(0), n(0), order(0), state(0) {}
};

OrganData apex_init_state;

/* Module definitions */

module A(OrganData); // apex
module I(OrganData); // internode
module L(OrganData); // leaf
module P(OrganData); // petiole
module K(OrganData); // flower
module C(OrganData); // petal
module D(OrganData); // seed pod
module B(OrganData); // axillary bud
module M(OrganData); // metamer for producing the above modules
module VisualAid(float);

/* Global variables */

float sim_time = 0.;

/* Organization of computation into phases */

Start: {
    sim_time = 0.;
    UseView(view001);
    UseView(view002);
    UseView(view003);
    UseView(view004);
}

EndEach: {
    // stop the simulation at 39 days to match the plant in the background image
    if (sim_time <= 39.)
        sim_time += DT;
    else
        Stop();
}

derivation length: STEPS;

Axiom: SB RollL(30) Right(-10.0)
          SetWidth(0.05) SetColor(1)
          SetElasticity(0,0.15)
          A(apex_init_state)
       EB
       VisualAid(0.);

production:

A(od) : {
	od.age += DT;
	produce A(od);
}
B(od) : {
	od.age += DT;
	produce B(od);
}
I(od) : {
	od.age += DT;
	produce I(od);
}
L(od) : {
	od.age += DT;
	produce L(od);
}
P(od) : {
	od.age += DT;
	produce P(od);
}
K(od) : {
	od.age += DT;
	produce K(od);
}
D(od) : {
	od.age += DT;
	produce D(od);
}

VisualAid(age) : {
    age += DT;
    produce VisualAid(age);
}

decomposition:
maximum depth: 200;

// define apical states
#define vegetative 0
#define reproductive 1
#define inactive 2

// define the apex behaviour
A(od) : {
	// model change of apex state
	if (od.state == vegetative && od.n >= val(VegPhytomers))
		od.state = reproductive;
	else if (od.state == reproductive && od.n >= val(MaxPhytomers))
		od.state = inactive;
    // check if apex should produce a phytomer
	if (od.state != inactive) { 
        // calculate plastochron given apex state
        float p;
        if (od.state == vegetative)
            // NOTE: the third term, (1.-val(ShortV)*od.order), was set to one in the paper,
            // but is here to give extra control over the vegetative plastochron
            p = val(VegPlastochron) * plastochron(od.n) * (1.-val(ShortV)*od.order); 
        else
            p = val(RepPlastochron);
        // if plastochron has elapsed
        if (od.age >= p) {
            od.age = od.age - p;
            nproduce M(od) // produce a phytomer
                     RollL(137.5); // simulate phyllotaxis
            od.n += 1; // then increase the phytomer count
            produce A(od);
        }
	}
}

M(od) : {
    nproduce I(od); // an internode
	if (od.state == vegetative) {
        produce SB B(od) EB  // an axillary bud, and
                SB L(od) EB; // a leaf
	}
	else { // in reproductive state
		produce SB Down(55.) P(od) K(od) EB; // NOTE: In the paper, the pedicel angle, Down(55), was missing
	}
}

B(od) : {
    // ignore buds above maximum branch order
    if (od.order >= val(MaxOrder) || val(SHOW_BRANCHES) == 0) 
        produce;
    // calculate bud vigour
    float v = bud_vigour(od.n) * (1. - val(ShortV) * od.order);
	if (v > val(MinBudVigour)) {
        // set the branching angle and the shoot gravitropic
        // response before new lateral apex is produced
        nproduce Down(val(BranchAngle) * branch_angle(od.n))
                 SetElasticity(0,val(Elast) * branch_tropism(od.n));
        // set organ data to produce new lateral apex
		od.state = vegetative;
		od.age = val(BudDelay) * bud_delay(od.n);
		od.order = od.order + 1;
		od.n = std::floor(val(VegPhytomers) * bud_init_node(od.n));
		produce A(od);
	}
    // NOTE: produce B(od); was incorrectly specified in the paper
}

K(od): {
    sran(od.n); 
    if (od.age > val(PoddingAge)) {
        if (ran(1) < prob_podding(sim_time)) {
            od.age = 0; // set pod age to zero
            produce SetElasticity(0,0.05) D(od); // flower becomes a pod
        } else produce; // or flower aborts
    }
    // otherwise flower remains
    // NOTE: produce K(od); was incorrectly given in the paper
}

interpretation:
maximum depth: 200;

#define InternodeInterpretation() \
I(od) : { \
    /* calculate the scaling factor for shortening */ \
    float shorten = 1. - val(ShortInt) * od.order; \
    /* calculate scaled length */ \
    float len = val(IntLen) * int_target_len(od.n) * shorten * int_length(od.age); \
    if (len <= 0.f) \
        produce; \
    /* calculate scaled radius */ \
    float rad = std::pow(len,val(ExpIntRad)); \
    int color = od.n >= val(VegPhytomers) ? 4 : 3; \
    /* draw a cylinder */ \
	produce SetColor(color) SetWidth(rad) F(len); \
}

#define LeafInterpretation() \
L(od) : { \
    if (val(SHOW_LEAVES)==0)\
        produce;\
    /* calculate leaf length */ \
	float len = val(LeafLen) * leaf_target_len(od.n) \
        * (1. - val(ShortLeaf) * od.order) * leaf_length(od.age); \
    if (len <= 0.f) \
        produce; \
    /* precompute leaf width scaling factor */ \
	float wid = val(LeafWid) * leaf_target_wid(od.n) * std::pow(len,val(ExpLeafWid)); \
    /* set leaf width at the base */ \
    /* set leaf insertion angle */ \
    /* set leaf cross-section to curve with index 1 */ \
    /* start drawing the generalized cylinder */ \
	nproduce SetElasticity(0,0) SetColor(7) \
            SetWidth(wid * leaf_width(0)) \
		    Down(val(BranchAngle) * branch_angle(od.n) * leaf_inclination(od.age)) \
		    CurrentContour(1) \
            StartGC(); \
    float x; \
    for (x = 0.; x < 1.-DX; x += DX) { \
        /* generate leaf segment */ \
        /* set width and orientation of next segment */ \
        nproduce F(len*DX) \
                 SetWidth(wid * leaf_width(x)) \
                 Down(leaf_bend(x) * leaf_angle(od.n) * DX); \
    } \
    nproduce SetWidth(wid * leaf_width(1.)) F(len*(1.-x)); /* generate last segment */ \
    produce EndGC(); /* end the generalized cylinder */ \
}

#define PedicleInterpretation() \
P(od) : { \
    if (val(SHOW_FLOWERS)==0)\
        produce;\
    /* calculate scaled length */ \
	float len = val(PedicelLen) * std::pow(int_target_len(od.n),val(ExpPedicelLen)) * pedicel_length(od.age); \
    /* calculate radius */ \
    float rad = std::pow(len,val(ExpPedicelRad)); \
    /* draw a cylinder */ \
	produce SetWidth(rad) F(len); \
}

#define FlowerInterpretation() \
K(od) : { \
    if (val(SHOW_FLOWERS)==0)\
        produce;\
    nproduce SetElasticity(0,0) SetColor(7); \
    /* draw corolla (i.e., four petals) */ \
    for (int i = 0; i < 4; i++) { \
        /* rotate in 90-degree increments */ \
        nproduce RollL(90. * float(i)) SB C(od) EB; \
    } \
    produce; \
} \
C(od): { \
    /* simulate flower opening */ \
    nproduce Down(180. * petal_angle(od.age)); \
    /* calculate petal length */ \
    float len = val(PetalLen) * std::pow(int_target_len(od.n),val(ExpPetalLen)) * petal_length(od.age); \
    /* calculate and set the initial petal width */ \
    float wid = std::pow(len,val(ExpPetalWid)); \
    nproduce SetWidth(wid * petal_width(0)); \
    /* set petal cross-section to curve with index 2 */ \
    nproduce CurrentContour(2) SetColor(5); \
    /* start drawing the generalized cylinder */ \
    nproduce StartGC(); \
    float x; \
    for (x = 0.; x < 1.-DX2; x += DX2) { \
        /* generate petal segment */ \
        nproduce F(len*DX2) SetWidth(wid * petal_width(x)); \
    } \
    nproduce SetWidth(wid * petal_width(1.)) F(len*(1.-x)); /* generate last segment */ \
    produce EndGC(); /* end the generalized cylinder */\
}

#define SiliqueInterpretation() \
D(od) : { \
    if (val(SHOW_FLOWERS)==0)\
        produce;\
    /* calculate pod length and radius */ \
    float len = val(PodLen) * std::pow(int_target_len(od.n),val(ExpPodLen)) * pod_length(od.age); \
    float rad = std::pow(len,val(ExpPodRad)); \
    /* draw main segment of the pod followed by the beak */ \
	produce SetColor(5) SB \
		   SetWidth(rad) StartGC F(len*0.8) \
		   SetWidth(0.1*rad) F(len*0.2) EndGC \
		   EB; \
}

// Draw image in the background
#define ViualAidInterpretation(view) \
VisualAid(age) : { \
    static int NUM_OTHER_TEXS = 1; \
    const float SF = 3.; \
	float w = 614.0 * SF; \
	float h = 1028.0 * SF; \
    int tex = 0; \
    nproduce SetElasticity(0,0.); \
    float move_back = 1.; \
    if (view == 3) { \
        move_back = 0.; \
        nproduce SetColor(2); \
    } \
    else { \
        move_back = 800.0; \
	    nproduce SetColor(1); \
    } \
    nproduce SetWidth(w); \
    nproduce CurrentContour(3); \
    if (view == 2 || view == 3) { \
        float topScale = 0.6; \
    	w = 1227.0 * SF; \
        nproduce CurrentTexture(tex+2+NUM_OTHER_TEXS); \
        nproduce TextureVCoeff(1.0/(h*topScale)) SetWidth(w*topScale);\
        nproduce SB MoveTo(0.,-250.,0.) RollL(90) Down(90) f(-h*0.5*topScale) \
                 StartGC() F(h*topScale) EndGC() EB; \
    } \
    if (view == 1 || view == 3) { \
        nproduce CurrentTexture(tex+1+NUM_OTHER_TEXS) TextureVCoeff(1.0/h); \
        nproduce SB MoveTo(-move_back,-150.*SF,-5.*SF) RollL(90) StartGC() F(h) EndGC() EB; \
    } \
    if (view == 0 || view == 3) { \
        nproduce CurrentTexture(tex+0+NUM_OTHER_TEXS) TextureVCoeff(1.0/h); \
        nproduce SB MoveTo(-5.*SF,-150.*SF,move_back) StartGC() F(h) EndGC() EB; \
    } \
    produce; \
}

vgroup 0:

vgroup view001:

InternodeInterpretation()
LeafInterpretation()
PedicleInterpretation()
FlowerInterpretation()
SiliqueInterpretation()

ViualAidInterpretation(0)

vgroup view002:

InternodeInterpretation()
LeafInterpretation()
PedicleInterpretation()
FlowerInterpretation()
SiliqueInterpretation()

ViualAidInterpretation(1)

vgroup view003:

InternodeInterpretation()
LeafInterpretation()
PedicleInterpretation()
FlowerInterpretation()
SiliqueInterpretation()

ViualAidInterpretation(2) 

vgroup view004:

InternodeInterpretation()
LeafInterpretation()
PedicleInterpretation()
FlowerInterpretation()
SiliqueInterpretation()

ViualAidInterpretation(3) 

